from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING

import requests
from pydantic import BaseModel, Field, SecretStr, ConfigDict

if TYPE_CHECKING:
    from bfabric import BfabricClientConfig


class TokenData(BaseModel):
    """Parsed token data from the B-Fabric token validation endpoint."""

    model_config = ConfigDict(
        populate_by_name=True, str_strip_whitespace=True, json_encoders={datetime: lambda v: v.isoformat()}
    )

    job_id: int = Field(alias="jobId")
    application_id: int = Field(alias="applicationId")

    entity_class: str = Field(alias="entityClassName")
    entity_id: int = Field(alias="entityId")

    user: str = Field(alias="user")
    user_ws_password: SecretStr = Field(alias="userWsPassword")

    token_expires: datetime = Field(alias="expiryDateTime")
    environment: str


def get_token_data(client_config: BfabricClientConfig, token: str) -> TokenData:
    """Returns the token data for the provided token.

    If the request fails, an exception is raised.
    """
    url = f"{client_config.base_url}/rest/token/validate"
    response = requests.get(url, params={"token": token})
    if not response.ok:
        response.raise_for_status()
    parsed = response.json()
    return TokenData.model_validate(parsed)



from __future__ import annotations

from typing import Any, TYPE_CHECKING

from suds.sax.text import Text
from suds.sudsobject import asdict

if TYPE_CHECKING:
    Value = list["Value"] | dict[str, "Value"] | str | int | float | bool | None | Any


def convert_suds_type(item: Any) -> int | str | Any:
    """
    Converts the suds type to an equivalent python type. There is, to my knowledge, only a single suds type which
    is currently ever return, namely 'Text'. Integers and doubles are already cast to their python equivalents and
    thus do not need to be explicitly handled. This may be subject to change in future versions
    :param item: The suds item
    :return: The item as a built-in python type
    """
    if isinstance(item, Text):
        return str(item)
    return item


def suds_asdict_recursive(d: Any, convert_types: bool = False) -> dict[str, Value]:
    """Convert Suds object into serializable format.
    https://stackoverflow.com/a/15678861
    :param d: The input suds object
    :param convert_types: A boolean to determine if the simple types return should be cast to python types
    :return: The suds object converted to an OrderedDict
    """
    out: dict[str, Value] = {}
    for k, v in asdict(d).items():
        if hasattr(v, "__keylist__"):
            out[k] = suds_asdict_recursive(v, convert_types=convert_types)
        elif isinstance(v, list):
            items: list[Value] = []
            for item in v:
                if hasattr(item, "__keylist__"):
                    items.append(suds_asdict_recursive(item, convert_types=convert_types))
                else:
                    items.append(convert_suds_type(item) if convert_types else item)
            out[k] = items
        else:
            out[k] = convert_suds_type(v) if convert_types else v
    return out



from __future__ import annotations

from typing import TYPE_CHECKING

import yaml
from loguru import logger

from bfabric.experimental import MultiQuery
from bfabric.experimental.entity_lookup_cache import EntityLookupCache

if TYPE_CHECKING:
    from pathlib import Path
    from collections.abc import Iterable
    from bfabric import Bfabric
    from typing import Any, Self


class Entity:
    ENDPOINT: str = ""

    def __init__(self, data_dict: dict[str, Any], client: Bfabric | None) -> None:
        self.__data_dict = data_dict
        self.__client = client

    @property
    def id(self) -> int:
        """Returns the entity's ID."""
        return int(self.__data_dict["id"])

    @property
    def web_url(self) -> str:
        if self._client is None:
            msg = "Cannot generate a web URL without a client's config information."
            raise ValueError(msg)
        return f"{self._client.config.base_url}/{self.ENDPOINT}/show.html?id={self.id}"

    @property
    def data_dict(self) -> dict[str, Any]:
        """Returns a shallow copy of the entity's data dictionary."""
        return self.__data_dict.copy()

    @property
    def _client(self) -> Bfabric | None:
        """Returns the client associated with the entity."""
        return self.__client

    @classmethod
    def find(cls, id: int, client: Bfabric) -> Self | None:
        """Finds an entity by its ID, if it does not exist `None` is returned."""
        cache = EntityLookupCache.instance()
        if cache and cache.contains(entity_type=cls, entity_id=id):
            return cache.get(entity_type=cls, entity_id=id)
        else:
            result = client.read(cls.ENDPOINT, obj={"id": int(id)})
            entity = cls(result[0], client=client) if len(result) == 1 else None
            if cache:
                cache.put(entity_type=cls, entity_id=id, entity=entity)
            return entity

    @classmethod
    def find_all(cls, ids: list[int], client: Bfabric) -> dict[int, Self]:
        """Returns a dictionary of entities with the given IDs. The order will generally match the input, however,
        if some entities are not found they will be omitted and a warning will be logged.
        """
        cache = EntityLookupCache.instance()
        ids_requested = cls.__check_ids_list(ids)

        # retrieve entities from cache and from B-Fabric as needed
        results_cached = cache.get_all(entity_type=cls, entity_ids=ids) if cache else {}
        results_fresh = cls.__retrieve_entities(
            client=client, ids_requested=ids_requested, ids_cached=results_cached.keys()
        )

        if cache:
            for entity_id, entity in results_fresh.items():
                cache.put(entity_type=cls, entity_id=entity_id, entity=entity)

        return cls.__ensure_results_order(ids_requested, results_cached, results_fresh)

    @classmethod
    def find_by(cls, obj: dict[str, Any], client: Bfabric, max_results: int | None = 100) -> dict[int, Self]:
        """Returns a dictionary of entities that match the given query."""
        result = client.read(cls.ENDPOINT, obj=obj, max_results=max_results)
        return {x["id"]: cls(x, client=client) for x in result}

    def dump_yaml(self, path: Path) -> None:
        """Writes the entity's data dictionary to a YAML file."""
        with path.open("w") as file:
            yaml.safe_dump(self.__data_dict, file)

    @classmethod
    def load_yaml(cls, path: Path, client: Bfabric | None) -> Self:
        """Loads an entity from a YAML file."""
        with path.open("r") as file:
            data = yaml.safe_load(file)
        return cls(data, client=client)

    def __contains__(self, key: str) -> Any:
        """Checks if a key is present in the data dictionary."""
        return key in self.__data_dict

    def __getitem__(self, key: str) -> Any:
        """Returns the value of a key in the data dictionary."""
        return self.__data_dict[key]

    def get(self, key: str, default: Any = None) -> Any:
        """Returns the value of a key in the data dictionary, or a default value if the key is not present."""
        return self.__data_dict.get(key, default)

    def __lt__(self, other: Entity) -> bool:
        """Compares the entity with another entity based on their IDs."""
        if self.ENDPOINT != other.ENDPOINT:
            return NotImplemented
        return self.id < other.id

    def __repr__(self) -> str:
        """Returns the string representation of the workunit."""
        return f"{self.__class__.__name__}({repr(self.__data_dict)}, client={repr(self.__client)})"

    __str__ = __repr__

    @classmethod
    def __check_ids_list(cls, ids: list[int]) -> list[int]:
        """Converts the ids to a list of integers (if they are not already) and raises an error if this fails or
        there are duplicates."""
        ids_requested = [int(id) for id in ids]
        if len(ids_requested) != len(set(ids_requested)):
            duplicates = [item for item in set(ids_requested) if ids_requested.count(item) > 1]
            raise ValueError(f"Duplicate IDs are not allowed, duplicates: {duplicates}")
        return ids_requested

    @classmethod
    def __retrieve_entities(
        cls, client: Bfabric, ids_requested: list[int], ids_cached: Iterable[int]
    ) -> dict[int, Self]:
        """Retrieves entities from B-Fabric that are not already in the cache"""
        ids = list(set(ids_requested) - set(ids_cached))
        if ids:
            if len(ids) > 100:
                result = MultiQuery(client).read_multi(cls.ENDPOINT, {}, "id", ids)
            else:
                result = client.read(cls.ENDPOINT, obj={"id": ids})
            return {x["id"]: cls(x, client=client) for x in result}
        else:
            return {}

    @classmethod
    def __ensure_results_order(
        cls,
        ids_requested: list[int],
        results_cached: dict[int, Self],
        results_fresh: dict[int, Self],
    ) -> dict[int, Self]:
        """Ensures the results are in the same order as requested and prints a warning if some results are missing."""
        results = {**results_cached, **results_fresh}
        results = {entity_id: results[entity_id] for entity_id in ids_requested if entity_id in results}
        if len(results) != len(ids_requested):
            logger.warning(f"Only found {len(results)} out of {len(ids_requested)}.")
        return results



from __future__ import annotations

from copy import deepcopy

from bfabric.results.result_container import ResultContainer
from bfabric.utils.paginator import page_iter
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from bfabric.bfabric import Bfabric


class MultiQuery:
    """Some advanced functionality that supports paginating over a list of conditions that is larger than the 100
    conditions limit of the API.
    This functionality might eventually be merged into the main Bfabric class but will probably be subject to some
    breaking changes and is not as thoroughly tested as the main classes functionality.
    """

    def __init__(self, client: Bfabric) -> None:
        self._client = client

    # TODO: Is this scope sufficient? Is there ever more than one multi-query parameter, and/or not at the root of dict?
    def read_multi(
        self,
        endpoint: str,
        obj: dict,
        multi_query_key: str,
        multi_query_vals: list,
        return_id_only: bool = False,
    ) -> ResultContainer:
        """Performs a 1-parameter multi-query (there is 1 parameter that takes a list of values)
        Since the API only allows BFABRIC_QUERY_LIMIT queries per page, split the list into chunks before querying
        :param endpoint: endpoint
        :param obj: query dictionary
        :param multi_query_key:  key for which the multi-query is performed
        :param multi_query_vals: list of values for which the multi-query is performed
        :param return_id_only: whether to return only the ids of the objects
        :return: List of responses, packaged in the results container

        NOTE: It is assumed that there is only 1 response for each value.
        """
        # TODO add `check` parameter
        response_tot = ResultContainer([], total_pages_api=0)
        obj_extended = deepcopy(obj)  # Make a copy of the query, not to make edits to the argument

        # Iterate over request chunks that fit into a single API page
        for page_vals in page_iter(multi_query_vals):
            obj_extended[multi_query_key] = page_vals

            # TODO: Test what happens if there are multiple responses to each of the individual queries.
            #     * What would happen?
            #     * What would happen if total number of responses would exceed 100 now?
            #     * What would happen if we naively made a multi-query with more than 100 values? Would API paginate
            #       automatically? If yes, perhaps we don't need this method at all?
            # TODO: It is assumed that a user requesting multi_query always wants all of the pages. Can anybody think of
            #   exceptions to this? -> yes, when not reading by id but for instance matching a pattern, one might not
            #   be aware that they might accidentally pull the whole db, so it's better to have max_results here as well
            #   but it is not completely trivial to implement
            response_this = self._client.read(endpoint, obj_extended, max_results=None, return_id_only=return_id_only)
            response_tot.extend(response_this, reset_total_pages_api=True)

        return response_tot

    # NOTE: Save-multi method is likely useless. When saving multiple objects, they all have different fields.
    #    One option would be to provide a dataframe, but it might struggle with nested dicts
    #    Likely best solution is to not provide this method, and let users run a for-loop themselves.
    # def save_multi(self, endpoint: str, obj_lst: list, **kwargs) -> ResultContainer:
    #     response_tot = ResultContainer([], self.result_type, total_pages_api = 0)
    #
    #     # Iterate over request chunks that fit into a single API page
    #     for page_objs in page_iter(obj_lst):
    #         response_page = self.save(endpoint, page_objs, **kwargs)
    #         response_tot.extend(response_page, reset_total_pages_api=True
    #
    #     return response_tot

    def delete_multi(self, endpoint: str, id_list: list[int]) -> ResultContainer:
        """Deletes multiple objects from `endpoint` by their ids."""
        # TODO document and test error handling
        # TODO add `check` parameter
        response_tot = ResultContainer([], total_pages_api=0)

        if not id_list:
            print("Warning, empty list provided for deletion, ignoring")
            return response_tot

        # Iterate over request chunks that fit into a single API page
        for page_ids in page_iter(id_list):
            response_page = self._client.delete(endpoint, page_ids)
            response_tot.extend(response_page, reset_total_pages_api=True)

        return response_tot

    def exists_multi(self, endpoint: str, key: str, value: list[int | str] | int | str) -> bool | list[bool]:
        """
        :param endpoint:  endpoint
        :param key:       A key for the query (e.g. id or name)
        :param value:     A value or a list of values
        :return:          Return a single bool or a list of bools for each value
            For each value, test if a key with that value is found in the API.
        """
        is_scalar = isinstance(value, (int, str))
        if is_scalar:
            return self._client.exists(endpoint=endpoint, key=key, value=value, check=True)
        elif not isinstance(value, list):
            raise ValueError("Unexpected data type", type(value))

        # 1. Read data for this id
        results = self.read_multi(endpoint, {}, key, value)

        # 2. Extract all the ids for which there was a response
        result_vals = []
        for r in results.results:
            if key in r:
                result_vals += [r[key]]
            elif "_" + key in r:  # TODO: Remove this if SUDS bug is ever resolved
                result_vals += [r["_" + key]]

        # 3. For each of the requested ids, return true if there was a response and false if there was not
        return [val in result_vals for val in value]



from __future__ import annotations

from functools import cached_property
from pathlib import Path
from typing import Any, TYPE_CHECKING

import dateutil.parser

from bfabric.entities.core.entity import Entity
from bfabric.entities.core.has_container_mixin import HasContainerMixin
from bfabric.entities.core.has_many import HasMany
from bfabric.entities.core.has_one import HasOne

if TYPE_CHECKING:
    from bfabric import Bfabric
    from bfabric.entities.application import Application
    from bfabric.entities.dataset import Dataset
    from bfabric.entities.externaljob import ExternalJob
    from bfabric.entities.parameter import Parameter
    from bfabric.entities.resource import Resource


class Workunit(Entity, HasContainerMixin):
    """Immutable representation of a single workunit in B-Fabric.
    :param data_dict: The dictionary representation of the workunit.
    """

    ENDPOINT = "workunit"

    def __init__(self, data_dict: dict[str, Any], client: Bfabric | None = None) -> None:
        super().__init__(data_dict=data_dict, client=client)

    application: HasOne[Application] = HasOne(entity="Application", bfabric_field="application")
    parameters: HasMany[Parameter] = HasMany(entity="Parameter", bfabric_field="parameter", optional=True)
    resources: HasMany[Resource] = HasMany(entity="Resource", bfabric_field="resource", optional=True)
    input_resources: HasMany[Resource] = HasMany(entity="Resource", bfabric_field="inputresource", optional=True)
    input_dataset: HasOne[Dataset] = HasOne(entity="Dataset", bfabric_field="inputdataset", optional=True)
    external_jobs: HasMany[ExternalJob] = HasMany(entity="ExternalJob", bfabric_field="externaljob", optional=True)

    @cached_property
    def parameter_values(self) -> dict[str, Any]:
        return {p.key: p.value for p in self.parameters.list}

    @cached_property
    def store_output_folder(self) -> Path:
        """Relative path in the storage for the workunit output."""
        if self.application is None:
            raise ValueError("Cannot determine the storage path without an application.")
        if self.application.storage is None:
            raise ValueError("Cannot determine the storage path without an application storage configuration.")
        date = dateutil.parser.parse(self.data_dict["created"])
        return Path(
            f"{self.application.storage['projectfolderprefix']}{self.container.id}",
            "bfabric",
            self.application["technology"].replace(" ", "_"),
            self.application["name"].replace(" ", "_"),
            date.strftime("%Y/%Y-%m/%Y-%m-%d/"),
            f"workunit_{self.id}",
        )



from __future__ import annotations

import tempfile
from pathlib import Path
from typing import Any, TYPE_CHECKING

from polars import DataFrame

from bfabric.entities.core.entity import Entity

if TYPE_CHECKING:
    from bfabric import Bfabric


class Dataset(Entity):
    """Immutable representation of a single dataset in B-Fabric.
    :param data_dict: The dictionary representation of the dataset.
    """

    ENDPOINT: str = "dataset"

    def __init__(self, data_dict: dict[str, Any], client: Bfabric | None = None) -> None:
        super().__init__(data_dict=data_dict, client=client)

    @property
    def column_names(self) -> list[str]:
        """Returns a list of column names in the dataset."""
        return [attr["name"] for attr in sorted(self.data_dict["attribute"], key=lambda a: int(a["position"]))]

    @property
    def column_types(self) -> dict[str, str]:
        """Returns a dictionary mapping column names to their data types."""
        return {
            attr["name"]: attr["type"] for attr in sorted(self.data_dict["attribute"], key=lambda a: int(a["position"]))
        }

    def to_polars(self) -> DataFrame:
        """Returns a Polars DataFrame representation of the dataset."""
        column_names = self.column_names
        data = []
        for item in self.data_dict["item"]:
            row_values = [x.get("value") for x in sorted(item["field"], key=lambda x: int(x["attributeposition"]))]
            data.append(dict(zip(column_names, row_values)))
        return DataFrame(data)

    def write_csv(self, path: Path, separator: str = ",") -> None:
        """Writes the dataset to a csv file at `path`, using the specified column `separator`."""
        self.to_polars().write_csv(path, separator=separator)

    def get_csv(self, separator: str = ",") -> str:
        """Returns the dataset as a csv string, using the specified column `separator`."""
        with tempfile.NamedTemporaryFile() as tmp_file:
            self.write_csv(Path(tmp_file.name), separator=separator)
            tmp_file.flush()
            tmp_file.seek(0)
            return tmp_file.read().decode()



from __future__ import annotations

import os
from typing import Annotated, Any

from loguru import logger
from pydantic import BaseModel, Field, model_validator
from pydantic_core import PydanticCustomError

from bfabric.config import BfabricAuth
from bfabric.config import BfabricClientConfig


class GeneralConfig(BaseModel):
    default_config: Annotated[str, Field(min_length=1)]


class EnvironmentConfig(BaseModel):
    config: BfabricClientConfig
    auth: BfabricAuth | None = None

    @model_validator(mode="before")
    @classmethod
    def gather_config(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Gathers all configs into the config field."""
        if not isinstance(values, dict):
            return values
        values["config"] = {key: value for key, value in values.items() if key not in ["login", "password"]}
        return values

    @model_validator(mode="before")
    @classmethod
    def gather_auth(cls, values: dict[str, Any]) -> dict[str, Any]:
        if isinstance(values, dict) and "login" in values:
            values["auth"] = BfabricAuth.model_validate(values)
        return values


class ConfigFile(BaseModel):
    general: Annotated[GeneralConfig, Field(alias="GENERAL")]
    environments: dict[str, EnvironmentConfig]

    @model_validator(mode="before")
    @classmethod
    def gather_configs(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Gathers all configs into the configs field."""
        configs = {}
        for key, value in values.items():
            if key != "GENERAL":
                configs[key] = value
        values["environments"] = configs
        return values

    @model_validator(mode="after")
    def validate_default_config(self) -> ConfigFile:
        """Validates that the default config is specified and is available in the configs."""
        if self.general.default_config not in self.environments:
            raise PydanticCustomError(
                "default_config_not_available",
                "Default config {default_config} not found in {available_configs}",
                {
                    "default_config": self.general.default_config,
                    "available_configs": set(self.environments.keys()),
                },
            )
        return self

    def get_selected_config_env(self, explicit_config_env: str | None) -> str:
        """Returns the name of the selected configuration, by checking the hierarchy of config_env definitions.
        1. If explicit_config_env is provided, it is used.
        2. If not, secondly, the parser will check if the environment variable `BFABRICPY_CONFIG_ENV` is declared
        3. If not, finally, the parser will select the default_config specified in GENERAL of the .bfabricpy.yml file
        """
        if explicit_config_env:
            return explicit_config_env
        elif "BFABRICPY_CONFIG_ENV" in os.environ:
            logger.debug(f"found BFABRICPY_CONFIG_ENV = {os.environ['BFABRICPY_CONFIG_ENV']}")
            return os.environ["BFABRICPY_CONFIG_ENV"]
        else:
            logger.debug(f"BFABRICPY_CONFIG_ENV not found, using default environment {self.general.default_config}")
            return self.general.default_config

    def get_selected_config(self, explicit_config_env: str | None = None) -> EnvironmentConfig:
        """Returns the selected configuration, by checking the hierarchy of config_env definitions.
        See selected_config_env for details."""
        return self.environments[self.get_selected_config_env(explicit_config_env=explicit_config_env)]



from collections import OrderedDict
from contextlib import redirect_stdout
from copy import deepcopy

import suds
import zeep

from bfabric import BfabricAuth, BfabricClientConfig
from bfabric.bfabric import get_system_auth
from bfabric.results.response_format_dict import drop_empty_elements, map_element_keys
from bfabric.engine.response_format_suds import suds_asdict_recursive

"""
This file is intended to eventually become a test to compare that Zeep and SUDS produce
the same or at least comparable output for the same requests. Important features
* Test if raw XML matches
* Test if parsed response (asdict) matches
For both, it is important to test
* different endpoints (user, run, ...)
* single match queries (e.g. {id=5}) vs multi match queries (e.g. {})

Observations:
* SUDS produces underscores in front of 'id', 'projectid' and 'classname'. Reasons currently unknown, may also affect
    other keywords. Currently, we remove underscores by explicitly providing keywords which to purge
* ZEEP does not match XML
    - Zeep generates additional keywords not present in XML, all of them have values None or empty list
    - Zeep misses some important keywords like 'id' and 'projectid' inside of nested XML, such as user->project. This
        behaviour is inconsistent, and only affects a fraction of users.
"""


def read_zeep(wsdl, fullQuery, raw=True):
    client = zeep.Client(wsdl)
    with client.settings(strict=False, raw_response=raw):
        ret = client.service.read(fullQuery)
    if raw:
        return ret.content
    else:
        return dict(zeep.helpers.serialize_object(ret, target_cls=dict))


def read_suds(wsdl, fullQuery, raw=True):
    client = suds.client.Client(wsdl, cache=None, retxml=raw)
    ret = client.service.read(fullQuery)
    if raw:
        return ret
    else:
        return suds_asdict_recursive(ret, convert_types=True)


def full_query(auth: BfabricAuth, query: dict, includedeletableupdateable: bool = False) -> dict:
    thisQuery = deepcopy(query)
    thisQuery["includedeletableupdateable"] = includedeletableupdateable

    return {"login": auth.login, "password": auth.password.get_secret_value(), "query": thisQuery}


def calc_both(
    auth: BfabricAuth,
    config: BfabricClientConfig,
    endpoint: str,
    query: dict,
    raw: bool = True,
):
    wsdl = "".join((config.base_url, "/", endpoint, "?wsdl"))
    fullQuery = full_query(auth, query)
    retZeep = read_zeep(wsdl, fullQuery, raw=raw)
    retSuds = read_suds(wsdl, fullQuery, raw=raw)
    return retZeep, retSuds


######################
# Raw XML tests
######################


def raw_test(auth: BfabricAuth, config: BfabricClientConfig, endpoint, query) -> None:
    print("Testing raw XML match for", endpoint, query)
    retZeep, retSuds = calc_both(auth, config, endpoint, query, raw=True)
    assert len(retZeep) == len(retSuds)
    assert retZeep == retSuds
    print("-- passed --")


config, auth = get_system_auth(config_env="TEST")
# raw_test(auth, config, 'user', {'id': 9026})
# raw_test(auth, config, 'user', {})

# root = etree.fromstring(retZeep)
# print(etree.tostring(root, pretty_print=True).decode())
# pprint(retZeepDict['user'][0]['order'])


######################
# Parsed dict comparison
######################


# Find the set of all basic types used in the nested container (made of dicts, ordered dicts and lists)
def recursive_get_types(generic_container) -> set:
    if isinstance(generic_container, (dict, OrderedDict)):
        type_sets_lst = [recursive_get_types(v) for k, v in generic_container.items()]
        return set().union(*type_sets_lst)
    elif isinstance(generic_container, list):
        type_sets_lst = [recursive_get_types(el) for el in generic_container]
        return set().union(*type_sets_lst)
    else:
        return {type(generic_container)}


def basic_data_type_match_test(auth, config, endpoint, query) -> None:
    print("Testing data types for", endpoint, query)
    retZeepDict, retSudsDict = calc_both(auth, config, endpoint, query, raw=False)
    typesZeep = recursive_get_types(retZeepDict)
    typesSuds = recursive_get_types(retZeepDict)
    print("Zeep", typesZeep)
    print("Suds", typesSuds)


# basic_data_type_match_test(auth, config, 'user', {'id': 9026})
# basic_data_type_match_test(auth, config, 'user', {})


# Compare two dictionaries/lists recursively. Print every time there is a discrepancy
def recursive_comparison(generic_container1, generic_container2, prefix: list) -> bool:
    matched = True

    if type(generic_container1) != type(generic_container2):
        print(prefix, "type", type(generic_container1), "!=", type(generic_container2))
        return False
    if isinstance(generic_container1, dict):
        allKeys = set(list(generic_container1.keys()) + list(generic_container2.keys()))
        for k in allKeys:
            if k not in generic_container1:
                print(prefix, "Not in 1: ", k, "=", generic_container2[k])
                print("- 1:", generic_container1)
                print("- 2:", generic_container2)
                matched = False
            elif k not in generic_container2:
                print(prefix, "Not in 2: ", k, "=", generic_container1[k])
                matched = False
            else:
                matched_recursive = recursive_comparison(generic_container1[k], generic_container2[k], prefix + [k])
                matched = matched and matched_recursive
    elif isinstance(generic_container1, list):
        if len(generic_container1) != len(generic_container2):
            print(prefix, "length", len(generic_container1), "!=", len(generic_container2))
            matched = False
        else:
            for i, (el1, el2) in enumerate(zip(generic_container1, generic_container2)):
                matched_recursive = recursive_comparison(el1, el2, prefix + [i])
                matched = matched and matched_recursive
    else:
        if generic_container1 != generic_container2:
            print(prefix, "value", generic_container1, "!=", generic_container2)
            matched = False

    return matched


def parsed_data_match_test(
    auth,
    config,
    endpoint,
    query,
    drop_empty: bool = True,
    drop_underscores_suds: bool = True,
    log_file_path: str = None,
) -> None:
    print("Testing parsed data match for", endpoint, query)
    retZeepDict, retSudsDict = calc_both(auth, config, endpoint, query, raw=False)

    if drop_empty:
        drop_empty_elements(retZeepDict, inplace=True)
        drop_empty_elements(retSudsDict, inplace=True)

    if drop_underscores_suds:
        map_element_keys(
            retSudsDict,
            {"_id": "id", "_classname": "classname", "_projectid": "projectid"},
            inplace=True,
        )

    if log_file_path is not None:
        with open(log_file_path, "w") as f, redirect_stdout(f):
            matched = recursive_comparison(retZeepDict, retSudsDict, prefix=[])
    else:
        matched = recursive_comparison(retZeepDict, retSudsDict, prefix=[])

    if matched:
        print("-- passed --")
    else:
        print("-- failed --")


parsed_data_match_test(
    auth,
    config,
    "user",
    {"id": 9026},
    drop_empty=True,
    drop_underscores_suds=True,
    log_file_path=None,
)
#
# parsed_data_match_test(auth, config, 'user', {}, drop_empty=True, drop_underscores_suds=True,
#                        log_file_path=None)

# parsed_data_match_test(auth, config, 'run', {}, drop_empty=True, drop_underscores_suds=True,
#                        log_file_path=None)

# print("Zeep", retZeep['user'][0]['project'][0])
# print("Suds", retSuds['user'][0]['project'][0])

# print("Zeep", retZeep['user'][0])
# print("Suds", retSuds['user'][0])



#! /usr/bin/env python
"""
Interfaces to the Sun/Oracle/Open Grid Engine batch queueing systems.
taken from https://vm-mad.googlecode.com/svn/trunk

https://arxiv.org/abs/1302.2529
DOI: 10.1007/978-3-642-38750-0_34

$Id: gridengine.py 2463 2016-09-23 14:55:50Z cpanse $
$HeadURL: http://fgcz-svn.uzh.ch/repos/scripts/trunk/linux/bfabric/apps/python/bfabric/gridengine.py $
$Date: 2016-09-23 16:55:50 +0200 (Fri, 23 Sep 2016) $
$Revision: 2463 $
"""

# Copyright (C) 2011, 2012 ETH Zurich and University of Zurich. All rights reserved.
# Copyright (C) 2014 Functional Genomics Center Zurich ETHZ|UZH. All rights reserved.
#
# Authors:
#   Christian Panse <cp@fgcz.ethz.ch>
#   Riccardo Murri <riccardo.murri@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# $Rev: 2463 $
# $HeadURL: http://fgcz-svn.uzh.ch/repos/scripts/trunk/linux/bfabric/apps/python/bfabric/gridengine.py $
# $Date: 2016-09-23 16:55:50 +0200 (Fri, 23 Sep 2016) $
# $Author: cpanse $

__docformat__ = "reStructuredText"
__version__ = "$Revision: 2463 $"


import os
import subprocess


class GridEngine:
    """
    interface to Open Grid Sceduler qsub
    """

    def __init__(
        self,
        user="*",
        queue="PRX@fgcz-r-035",
        GRIDENGINEROOT="/export/bfabric/bfabric/",
    ):
        """
        Set up parameters for querying Grid Engine.

        SGEROOT is essential.
        """

        self.user = user
        self.queue = queue
        self.qsubbin = f"{GRIDENGINEROOT}/bin/qsub"

        os.environ["SGE_ROOT"] = GRIDENGINEROOT

    def qsub(self, script, arguments=""):
        """
        if qsub and script are files do
        qsub as fire and forget

        todo: pass stderr and stdout file location as argument
        """
        qsub_cmd = [self.qsubbin, "-q", self.queue, script, " ".join(arguments)]

        if not os.path.isfile(self.qsubbin):
            print(f"{self.qsubbin} can not be found.")
            return

        if not os.path.isfile(script):
            print(f"'{script}' - no such file.")
            return

        try:
            qsub_process = subprocess.Popen(qsub_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=False)
            stdout, stderr = qsub_process.communicate()

            return stdout

        # except subprocess.CalledProcessError, ex:
        except:
            # logging.error("Error running '%s': '%s'; exit code %d", str.join(' ', qstat_cmd), stderr, ex.returncode)
            raise


def main():
    print("hello world!")
    pass


if __name__ == "__main__":
    main()



from __future__ import annotations

import base64
from pathlib import Path

import yaml
from loguru import logger

from bfabric.bfabric import Bfabric
from bfabric.entities import ExternalJob, Executable
from bfabric.wrapper_creator.slurm import SLURM


class BfabricSubmitter:
    """
    the class is used by the submitter which is executed by the bfabric system.
    """

    slurm_dict = {
        "MaxQuant_textfiles_sge": {
            "partition": "prx",
            "nodelist": "fgcz-r-033",
            "memory": "1G",
        },
        "fragpipe": {"partition": "prx", "nodelist": "fgcz-r-033", "memory": "256G"},
        "MaxQuant": {"partition": "maxquant", "nodelist": "fgcz-r-033", "memory": "4G"},
        "scaffold_generic": {
            "partition": "scaffold",
            "nodelist": "fgcz-r-033",
            "memory": "256G",
        },
        "MSstats dataProcess": {
            "partition": "prx",
            "nodelist": "fgcz-r-033",
            "memory": "64G",
        },
        "MaxQuant_sampleSizeEstimation": {
            "partition": "prx",
            "nodelist": "fgcz-r-028",
            "memory": "2G",
        },
        "ProteomeDiscovererQC": {
            "partition": "prx",
            "nodelist": "fgcz-r-035",
            "memory": "2G",
        },
    }

    def __init__(
        self,
        client: Bfabric,
        externaljobid: int,
        user: str = "*",
        partition: str = "prx",
        nodelist: str = "fgcz-r-028",
        memory: str = "10G",
        scheduleroot: str = "/export/bfabric/bfabric/",
        scheduler: str = "GridEngine",
        script_dir: Path = Path("/home/bfabric/prx"),
    ) -> None:
        self._client = client
        self._executable_file_list = []

        self.partition = partition
        self.nodelist = nodelist
        self.memory = memory
        self.scheduleroot = scheduleroot
        self.scheduler = scheduler
        self.user = user
        self._script_dir = script_dir

        self.external_job = ExternalJob.find(id=externaljobid, client=client)
        self.workunit = self.external_job.workunit
        self.parameters = self.workunit.parameter_values
        self.application = self.workunit.application

        default_config = self.slurm_dict.get(self.application["name"], {})
        self.partition = self.parameters.get("partition", default_config.get("partition"))
        self.nodelist = self.parameters.get("nodelist", default_config.get("nodelist"))
        self.memory = self.parameters.get("memory", default_config.get("memory"))

        logger.debug(f"partition={self.partition}")
        logger.debug(f"nodelist={self.nodelist}")
        logger.debug(f"memory={self.memory}")
        logger.debug("__init__ DONE")

    def submit_slurm(self, script: str = "/tmp/runme.bash") -> None:
        slurm = SLURM(slurm_root=self.scheduleroot)
        logger.debug(script)
        logger.debug(type(script))
        res_slurm_batch = slurm.sbatch(script=script)
        logger.debug(f"{res_slurm_batch}")

    def compose_bash_script(self, configuration=None, configuration_parser=lambda x: yaml.safe_load(x)) -> str:
        """
        composes the bash script which is executed by the submitter (sun grid engine).
        as an argument it takes a configuration file, e.g., yaml, xml, json, or whatsoever, and a parser function.

        it returns a str object containing the code.

        :rtype : str
        """

        # assert isinstance(configuration, str)
        config = configuration_parser(configuration)

        _cmd_template = """#!/bin/bash
# Maria d'Errico
# Christian Panse
# 2020-09-28
# 2020-09-29
# https://GitHub.com/fgcz/bfabricPy/
# Slurm
#SBATCH --partition={partition}
#SBATCH --nodelist={nodelist}
#SBATCH -n 1
#SBATCH -N 1
#SBATCH --cpus-per-task=1
#SBATCH --mem-per-cpu={memory}
#SBATCH -e {stderr_url}
#SBATCH -o {stdout_url}
#SBATCH --job-name=WU{workunit_id}
#SBATCH --chdir=/home/bfabric
#SBATCH --export=ALL,HOME=/home/bfabric

set -euxo pipefail

# diagnostics
id
hostname
echo "${{PATH}}"
echo "${{LD_LIBRARY_PATH:-}}"
echo "${{MODULEPATH:-}}"
echo "${{LOADEDMODULES:-}}"
echo "${{SLURM_JOB_ID:-}}"
echo "${{TMPDIR:-}}"
echo "${{PYTHONPATH:-}}"
echo "${{R_LIBS:-}}"

export EMAIL="{job_notification_emails}"
export EXTERNALJOB_ID={external_job_id}
export RESSOURCEID_OUTPUT={resource_id_output}
export RESSOURCEID_STDOUT_STDERR="{resource_id_stderr} {resource_id_stdout}"
export OUTPUT="{output_list}"
export WORKUNIT_ID="{workunit_id}"
TEMPDIR="/home/bfabric/prx"

_OUTPUT=`echo $OUTPUT | cut -d"," -f1`
test $? -eq 0 && _OUTPUTHOST=`echo $_OUTPUT | cut -d":" -f1`
test $? -eq 0 && _OUTPUTPATH=`echo $_OUTPUT | cut -d":" -f2`
test $? -eq 0 && _OUTPUTPATH=`dirname $_OUTPUTPATH`
test $? -eq 0 && ssh $_OUTPUTHOST "mkdir -p $_OUTPUTPATH"
test $? -eq 0 && echo $$ > $TEMPDIR/$$
test $? -eq 0 && scp $TEMPDIR/$$ $OUTPUT

if [ $? -eq 1 ];
then
    echo "writing to output url failed!";
    exit 1;
fi

# job configuration set by B-Fabrics wrapper_creator executable
# application parameter/configuration
cat > $TEMPDIR/config_WU$WORKUNIT_ID.yaml <<EOF
{configuration}
EOF



## interrupt here if you want to do a semi-automatic processing
if [ -x /usr/bin/mutt ];
then
    cat $0 > $TEMPDIR/$EXTERNALJOB_ID.bash

    (who am i; hostname; uptime; echo $0; pwd; ps;) \
    | mutt -s "WORKUNIT_ID=$WORKUNIT_ID EXTERNALJOB_ID=$EXTERNALJOB_ID" $EMAIL \
        -a $TEMPDIR/$EXTERNALJOB_ID.bash $TEMPDIR/config_WU$WORKUNIT_ID.yaml
fi
# exit 0

# Set the workunit status to processing
bfabric_setWorkunitStatus_processing.py $WORKUNIT_ID

# run the application
test -f $TEMPDIR/config_WU$WORKUNIT_ID.yaml && {executable} $TEMPDIR/config_WU$WORKUNIT_ID.yaml


if [ $? -eq 0 ];
then
     ssh fgcz-r-035.uzh.ch "bfabric_setResourceStatus_available.py $RESSOURCEID_OUTPUT" \
     | mutt -s "WORKUNIT_ID=$WORKUNIT_ID EXTERNALJOB_ID=$EXTERNALJOB_ID DONE" $EMAIL

     bfabric_save_workflowstep.py $WORKUNIT_ID
     bfabric_setExternalJobStatus_done.py $EXTERNALJOB_ID
     bfabric_setWorkunitStatus_available.py $WORKUNIT_ID
    echo $?
else
    echo "application failed"
     mutt -s "WORKUNIT_ID=$WORKUNIT_ID EXTERNALJOB_ID=$EXTERNALJOB_ID failed" $EMAIL < /dev/null
     bfabric_setResourceStatus_available.py $RESSOURCEID_STDOUT_STDERR $RESSOURCEID;
     bfabric_setWorkunitStatus_failed.py $WORKUNIT_ID
    exit 1;
fi

# should be available also as zero byte files
bfabric_setResourceStatus_available.py $RESSOURCEID_STDOUT_STDERR


exit 0
""".format(
            partition=self.partition,
            stderr_url=config["job_configuration"]["stderr"]["url"],
            stdout_url=config["job_configuration"]["stdout"]["url"],
            external_job_id=config["job_configuration"]["external_job_id"],
            resource_id_output=config["job_configuration"]["output"]["resource_id"],
            resource_id_stderr=config["job_configuration"]["stderr"]["resource_id"],
            resource_id_stdout=config["job_configuration"]["stdout"]["resource_id"],
            output_list=",".join(config["application"]["output"]),
            configuration=configuration,
            executable=config["job_configuration"]["executable"],
            workunit_id=config["job_configuration"]["workunit_id"],
            nodelist=self.nodelist,
            memory=self.memory,
            job_notification_emails=self._client.config.job_notification_emails,
        )

        return _cmd_template

    def submitter_yaml(self) -> None:
        """
        implements the default submitter

        the function fetches the yaml base64 configuration file linked to the external job id out of the B-Fabric
        system. Since the file can not be staged to the LRMS as argument, we copy the yaml file into the bash script
        and stage it on execution the application.

        TODO(cp): create the output url before the application is started.

        return None
        """
        executables = Executable.find_by({"workunitid": self.workunit.id}, client=self._client).values()
        for executable in executables:
            if not executable["base64"]:
                continue

            logger.debug(f"executable = {executable}")
            content = base64.b64decode(executable["base64"].encode()).decode()
            logger.debug(content)
            _cmd_template = self.compose_bash_script(
                configuration=content, configuration_parser=lambda x: yaml.safe_load(x)
            )

            bash_script_file = Path(
                self._script_dir,
                f"workunitid-{self.workunit.id}_externaljobid-{self.external_job.id}"
                f"_executableid-{self.external_job.executable.id}.bash",
            )

            bash_script_file.write_text(_cmd_template)

            if self.scheduler == "GridEngine":
                raise NotImplementedError
                # self.submit_gridengine(bash_script_file)
            else:
                self.submit_slurm(str(bash_script_file))
            self._executable_file_list.append(str(bash_script_file))

        self._client.save("externaljob", {"id": self.external_job.id, "status": "done"})

    def get_job_script(self) -> list[str]:
        return self._executable_file_list



from __future__ import annotations

from typing import TYPE_CHECKING

import polars as pl

if TYPE_CHECKING:
    from pathlib import Path
    from bfabric.bfabric import Bfabric


def polars_to_bfabric_type(dtype: pl.DataType) -> str | None:
    """Returns the B-Fabric type for a given Polars data type, defaulting to String if no correspondence is found."""
    if str(dtype).startswith("Int"):
        return "Integer"
    elif str(dtype).startswith("String"):
        return "String"
    else:
        return "String"


def polars_to_bfabric_dataset(
    data: pl.DataFrame,
) -> dict[str, list[dict[str, int | str | float]]]:
    """Converts a Polars DataFrame to a B-Fabric dataset representation."""
    attributes = [
        {
            "name": col,
            "position": i + 1,
            "type": polars_to_bfabric_type(data[col].dtype),
        }
        for i, col in enumerate(data.columns)
    ]
    items = [
        {
            "field": [{"attributeposition": i_field + 1, "value": value} for i_field, value in enumerate(row)],
            "position": i_row + 1,
        }
        for i_row, row in enumerate(data.iter_rows())
    ]
    return {"attribute": attributes, "item": items}


def bfabric_save_csv2dataset(
    client: Bfabric,
    csv_file: Path,
    dataset_name: str,
    container_id: int,
    workunit_id: int | None,
    sep: str,
    has_header: bool,
    invalid_characters: str,
) -> None:
    """Creates a dataset in B-Fabric from a csv file."""
    data = pl.read_csv(csv_file, separator=sep, has_header=has_header, infer_schema_length=None)
    check_for_invalid_characters(data=data, invalid_characters=invalid_characters)
    obj = polars_to_bfabric_dataset(data)
    obj["name"] = dataset_name
    obj["containerid"] = container_id
    if workunit_id is not None:
        obj["workunitid"] = workunit_id
    endpoint = "dataset"
    res = client.save(endpoint=endpoint, obj=obj)
    print(res.to_list_dict()[0])


def check_for_invalid_characters(data: pl.DataFrame, invalid_characters: str) -> None:
    """Raises a RuntimeError if any cell in the DataFrame contains an invalid character."""
    if not invalid_characters:
        return
    invalid_columns_df = data.select(pl.col(pl.String).str.contains_any(list(invalid_characters)).any())
    if invalid_columns_df.is_empty():
        return
    invalid_columns = (
        invalid_columns_df.transpose(include_header=True, header_name="column")
        .filter(pl.col("column_0"))["column"]
        .to_list()
    )
    if invalid_columns:
        raise RuntimeError(f"Invalid characters found in columns: {invalid_columns}")



from __future__ import annotations

from typing import TypeVar, TYPE_CHECKING

from bfabric.entities.core.relationship import Relationship

if TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    from bfabric.entities.core.entity import Entity

E = TypeVar("E", bound="Entity")
T = TypeVar("T")


class HasOne(Relationship[E]):
    def __init__(self, entity: str, *, bfabric_field: str, optional: bool = False) -> None:
        super().__init__(entity)
        self._bfabric_field = bfabric_field
        self._optional = optional

    def __get__(self, obj: T | None, objtype: type[T] | None = None) -> E | None:
        cache_attr = f"_HasOne__{self._bfabric_field}_cache"
        if not hasattr(obj, cache_attr):
            setattr(obj, cache_attr, self._load_entity(obj=obj))
        return getattr(obj, cache_attr)

    def _load_entity(self, obj: T) -> E | None:
        client = obj._client
        entity_data = obj.data_dict.get(self._bfabric_field)
        if self._optional and entity_data is None:
            return None
        elif entity_data is None:
            raise ValueError(f"Field {repr(self._bfabric_field)} is required")
        return self._entity_type.find(id=entity_data["id"], client=client)



from __future__ import annotations

from typing import Annotated

from pydantic import BaseModel, Field, SecretStr


class BfabricAuth(BaseModel):
    """Holds the authentication data for the B-Fabric client."""

    login: Annotated[str, Field(min_length=3)]
    password: Annotated[SecretStr, Field(min_length=32, max_length=32)]



from __future__ import annotations

from collections import defaultdict, OrderedDict
from contextlib import contextmanager
from typing import TypeVar, Generic, TYPE_CHECKING

from loguru import logger

if TYPE_CHECKING:
    from collections.abc import Generator
    from collections.abc import Hashable
    from bfabric.entities.core.entity import Entity  # type: ignore

T = TypeVar("T")
E = TypeVar("E", bound="Entity")


class Cache(Generic[T]):
    """A FIFO cache with a maximum size, implemented by an OrderedDict."""

    def __init__(self, max_size: int) -> None:
        self._entries: OrderedDict[Hashable, T] = OrderedDict()
        self._max_size = max_size

    def get(self, key: Hashable) -> T | None:
        """Returns the value with the given key, if it exists, and marks it as used.

        If the key does not exist, returns None.
        """
        if key in self._entries:
            self._entries.move_to_end(key)
            return self._entries[key]

    def put(self, key: Hashable, value: T) -> None:
        """Puts a key-value pair into the cache, marking it as used."""
        if self._max_size != 0 and len(self._entries) >= self._max_size:
            self._entries.popitem(last=False)
        self._entries[key] = value

    def __contains__(self, key: Hashable) -> bool:
        """Returns whether the cache contains a key."""
        return key in self._entries


class EntityLookupCache:
    """Implements the logic for caching entity lookup.

    :param max_size: The maximum size of the cache. If 0, the cache has no size limit.
    """

    __class_instance = None

    def __init__(self, max_size: int = 0) -> None:
        self._caches: dict[type[Entity], Cache[Entity | None]] = defaultdict(lambda: Cache(max_size=max_size))

    def contains(self, entity_type: type[Entity], entity_id: int) -> bool:
        """Returns whether the cache contains an entity with the given type and ID."""
        return entity_id in self._caches[entity_type]

    def get(self, entity_type: type[E], entity_id: int) -> E | None:
        """Returns the entity with the given type and ID, if it exists in the cache."""
        if self._caches[entity_type].get(entity_id):
            logger.debug(f"Cache hit for entity {entity_type} with ID {entity_id}")
            return self._caches[entity_type].get(entity_id)
        else:
            logger.debug(f"Cache miss for entity {entity_type} with ID {entity_id}")
            return None

    def get_all(self, entity_type: type[Entity], entity_ids: list[int]) -> dict[int, Entity]:
        """Returns a dictionary of entities with the given type and IDs,
        containing only the entities that exist in the cache.
        """
        results = {entity_id: self.get(entity_type, entity_id) for entity_id in entity_ids}
        return {entity_id: result for entity_id, result in results.items() if result is not None}

    def put(self, entity_type: type[Entity], entity_id: int, entity: Entity | None) -> None:
        """Puts an entity with the given type and ID into the cache."""
        logger.debug(f"Caching entity {entity_type} with ID {entity_id}")
        self._caches[entity_type].put(entity_id, entity)

    @classmethod
    @contextmanager
    def enable(cls, max_size: int = 0) -> Generator[None, None, None]:
        """Context manager that enables the EntityLookupCache singleton instance, i.e. every entity lookup by ID
        within this context will be cached. The cache is cleared after the context exits.
        """
        existing_cache = cls.__class_instance is not None
        if not existing_cache:
            cls.__class_instance = cls(max_size=max_size)
            # TODO what to do if existing_cache and max_size mismatch?
            # TODO another relevant use case could be selectively caching only some entities, whereas others should be
            #      reloaded
            # TODO finally, there is the question about persistent caches (e.g. storages do not change that often)
        try:
            yield
        finally:
            if not existing_cache:
                cls.__class_instance = None

    @classmethod
    def instance(cls) -> EntityLookupCache | None:
        """Returns the singleton instance of the EntityLookupCache."""
        return cls.__class_instance



from __future__ import annotations

from typing import Annotated, Any

from pydantic import BaseModel, BeforeValidator, Field, TypeAdapter, AnyHttpUrl

http_url_adapter = TypeAdapter(AnyHttpUrl)


class BfabricClientConfig(BaseModel):
    """Holds the configuration for the B-Fabric client for connecting to particular instance of B-Fabric.

    Parameters:
        base_url (optional): The API base url
        application_ids (optional): Map of application names to ids.
        job_notification_emails (optional): Space-separated list of email addresses to notify when a job finishes.
    """

    # TODO consider using AnyHttpUrl in the future directly and make mandatory
    base_url: Annotated[
        str,
        BeforeValidator(lambda value: str(http_url_adapter.validate_python(value))),
        Field(default="https://fgcz-bfabric.uzh.ch/bfabric"),
    ]
    application_ids: Annotated[dict[str, int], Field(default_factory=dict)]
    job_notification_emails: Annotated[str, Field(default="")]

    def __init__(self, **kwargs: Any) -> None:
        # TODO remove this custom constructor (note that this is currently used in some places when "None" is passed)
        super().__init__(**{key: value for key, value in kwargs.items() if value is not None})

    def copy_with(
        self,
        base_url: str | None = None,
        application_ids: dict[str, int] | None = None,
    ) -> BfabricClientConfig:
        """Returns a copy of the configuration with new values applied, if they are not None."""
        return BfabricClientConfig(
            base_url=base_url if base_url is not None else self.base_url,
            application_ids=(application_ids if application_ids is not None else self.application_ids),
            job_notification_emails=self.job_notification_emails,
        )

    def __str__(self) -> str:
        return repr(self)



from __future__ import annotations

import base64
from functools import cached_property
from typing import Any, TYPE_CHECKING

from bfabric.entities.core.entity import Entity
from bfabric.entities.core.has_one import HasOne

if TYPE_CHECKING:
    from bfabric import Bfabric


class Executable(Entity):
    ENDPOINT = "executable"

    storage = HasOne(entity="Storage", bfabric_field="storage", optional=True)

    def __init__(self, data_dict: dict[str, Any], client: Bfabric | None = None) -> None:
        super().__init__(data_dict=data_dict, client=client)

    @cached_property
    def decoded(self) -> str | None:
        if "base64" in self:
            return base64.decodebytes(self["base64"].encode("utf-8")).decode("utf-8")
        return None



from __future__ import annotations

from pathlib import Path

import yaml
from loguru import logger

from bfabric.config import BfabricAuth
from bfabric.config import BfabricClientConfig
from bfabric.config import ConfigFile


def read_config(
    config_path: str | Path,
    config_env: str | None = None,
) -> tuple[BfabricClientConfig, BfabricAuth | None]:
    """
    Reads bfabricpy.yml file, parses it, extracting authentication and configuration data
    :param config_path:   Path to the configuration file. It is assumed the file exists
    :param config_env:    Configuration environment to use. If not given, it is deduced.
    :return: Configuration and Authentication class instances

    NOTE: BFabricPy expects a .bfabricpy.yml of the format, as seen in bfabricPy/tests/unit/example_config.yml
    * The general field always has to be present
    * There may be any number of environments, with arbitrary names. Here, they are called PRODUCTION and TEST
    * Must specify correct login, password and base_url for each environment.
    * application and job_notification_emails fields are optional
    * The default environment will be selected as follows:
        - First, parser will check if the optional argument `config_env` is provided directly to the parser function
        - If not, secondly, the parser will check if the environment variable `BFABRICPY_CONFIG_ENV` is declared
        - If not, finally, the parser will select the default_config specified in [GENERAL] of the .bfabricpy.yml file
    """
    logger.debug(f"Reading configuration from: {config_path}")
    config_file = ConfigFile.model_validate(yaml.safe_load(Path(config_path).read_text()))
    env_config = config_file.get_selected_config(explicit_config_env=config_env)
    return env_config.config, env_config.auth



from .application import Application
from .dataset import Dataset
from .executable import Executable
from .externaljob import ExternalJob
from .multiplexid import MultiplexId
from .multiplexkit import MultiplexKit
from .order import Order
from .parameter import Parameter
from .project import Project
from .resource import Resource
from .sample import Sample
from .storage import Storage
from .workunit import Workunit

__all__ = [
    "Application",
    "Dataset",
    "Executable",
    "ExternalJob",
    "MultiplexId",
    "MultiplexKit",
    "Order",
    "Parameter",
    "Project",
    "Resource",
    "Sample",
    "Storage",
    "Workunit",
]



from __future__ import annotations

from typing import Generic, TypeVar, TYPE_CHECKING
from bfabric.entities.core.relationship import Relationship
from polars import DataFrame

if TYPE_CHECKING:
    from collections.abc import Iterator
    from bfabric import Bfabric

    # noinspection PyUnresolvedReferences
    from bfabric.entities.core.entity import Entity

E = TypeVar("E", bound="Entity")
T = TypeVar("T")


class HasMany(Relationship[E]):
    def __init__(
        self,
        entity: str,
        *,
        bfabric_field: str | None = None,
        ids_property: str | None = None,
        client_property: str = "_client",
        optional: bool = False,
    ) -> None:
        super().__init__(entity)
        self._bfabric_field = bfabric_field
        self._ids_property = ids_property
        self._client_property = client_property
        self._optional = optional

    def __get__(self, obj: T | None, objtype: type[T] | None = None) -> _HasManyProxy:
        cache_attr = f"_HasMany__{self._ids_property or self._bfabric_field}_cache"
        if obj is None:
            raise ValueError("Cannot access HasMany relationship on class")
        if not hasattr(obj, cache_attr):
            ids = self._get_ids(obj)
            client = getattr(obj, self._client_property)
            setattr(
                obj,
                cache_attr,
                _HasManyProxy(entity_type=self._entity_type, ids=ids, client=client),
            )
        return getattr(obj, cache_attr)

    def _get_ids(self, obj: T) -> list[int]:
        if self._bfabric_field is not None:
            if self._ids_property is not None:
                raise ValueError("Exactly one of bfabric_field and ids_property must be set, but both are set")
            if self._optional and self._bfabric_field not in obj.data_dict:
                return []
            return [x["id"] for x in obj.data_dict[self._bfabric_field]]
        elif self._ids_property is not None:
            if self._optional and not hasattr(obj, self._ids_property):
                return []
            return getattr(obj, self._ids_property)
        else:
            raise ValueError("Exactly one of bfabric_field and ids_property must be set, but neither is set")


class _HasManyProxy(Generic[E]):
    def __init__(self, entity_type: type[E], ids: list[int], client: Bfabric) -> None:
        self._entity_type = entity_type
        self._ids = ids
        self._client = client
        self._items: dict[int, E] = {}

    @property
    def ids(self) -> list[int]:
        return self._ids

    @property
    def list(self) -> list[E]:
        self._load_all()
        return sorted(self._items.values(), key=lambda x: self._items.keys())

    @property
    def polars(self) -> DataFrame:
        self._load_all()
        return DataFrame([x.data_dict for x in self._items.values()])

    def __getitem__(self, key: int) -> E:
        # TODO should key=0 map to first element etc instead? this can be a bit confusing at times
        self._load_all()
        return self._items[key]

    def __iter__(self) -> Iterator[E]:
        self._load_all()
        return iter(sorted(self._items.values(), key=lambda x: self._items.keys()))

    def _load_all(self) -> None:
        if not self._items and self._ids:
            self._items = self._entity_type.find_all(ids=self._ids, client=self._client)

    def __len__(self) -> int:
        return len(self._ids)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._entity_type}, {self._ids}, {self._client})"

    __str__ = __repr__



from __future__ import annotations

import math
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Generator

# Single page query limit for BFabric API (as of time of writing, adapt if it changes)
BFABRIC_QUERY_LIMIT = 100


def page_iter(objs: list, page_size: int = BFABRIC_QUERY_LIMIT) -> Generator[list, None, None]:
    """
    :param objs:       A list of objects to provide to bfabric as part of a query
    :param page_size:  Number of objects per page
    :return:           An iterator over chunks that would be sent to bfabric, 1 chunk per query
    """

    for i in range(0, len(objs), page_size):
        yield objs[i : i + page_size]


def compute_requested_pages(
    n_page_total: int,
    n_item_per_page: int,
    n_item_offset: int,
    n_item_return_max: int | None,
) -> tuple[list[int], int]:
    """Returns the page indices that need to be requested to get all requested items.
    :param n_page_total: Total number of pages available
    :param n_item_per_page: Number of items per page
    :param n_item_offset: Number of items to skip from the beginning
    :param n_item_return_max: Maximum number of items to return
    :return:
        - list of page indices that need to be requested
        - initial page offset (0-based), i.e. the i-th item from which onwards to retain results
    """
    # B-Fabric API uses 1-based indexing for pages
    index_start = 1

    # Determine the page indices to request
    # If n_item_return_max is not provided, we will return all items
    if n_item_return_max is None:
        n_item_return_max = n_page_total * n_item_per_page

    # Determine the page indices to request
    idx_max_return = math.ceil((n_item_return_max + n_item_offset) / n_item_per_page)
    idx_arr = [idx + index_start for idx in range(n_item_offset // n_item_per_page, min(n_page_total, idx_max_return))]

    # Determine the initial offset on the first page
    initial_offset = min(n_item_offset, n_item_return_max) % n_item_per_page

    return idx_arr, initial_offset



"""B-Fabric Application Interface using WSDL

Copyright (C) 2014 - 2024 Functional Genomics Center Zurich ETHZ|UZH. All rights reserved.

Licensed under GPL version 3

Authors:
  Marco Schmidt <marco.schmidt@fgcz.ethz.ch>
  Christian Panse <cp@fgcz.ethz.ch>
  Leonardo Schwarz
  Aleksejs Fomins
"""

from __future__ import annotations

import base64
import importlib.metadata
from contextlib import contextmanager
from datetime import datetime
from enum import Enum
from functools import cached_property
from pathlib import Path
from pprint import pprint
from typing import Literal, Any, TYPE_CHECKING

import sys
from loguru import logger
from rich.console import Console

from bfabric.bfabric_config import read_config
from bfabric.config import BfabricAuth
from bfabric.config import BfabricClientConfig
from bfabric.engine.engine_suds import EngineSUDS
from bfabric.engine.engine_zeep import EngineZeep
from bfabric.rest.token_data import get_token_data
from bfabric.results.result_container import ResultContainer
from bfabric.utils.cli_integration import DEFAULT_THEME, HostnameHighlighter
from bfabric.utils.paginator import compute_requested_pages, BFABRIC_QUERY_LIMIT

if TYPE_CHECKING:
    from collections.abc import Generator


class BfabricAPIEngineType(Enum):
    """Choice of engine to use."""

    SUDS = 1
    ZEEP = 2


class Bfabric:
    """Bfabric client class, providing general functionality for interaction with the B-Fabric API.
    Use `Bfabric.from_config` to create a new instance.
    :param config: Configuration object
    :param auth: Authentication object (if `None`, it has to be provided using the `with_auth` context manager)
    :param engine: Engine type to use for the API. Default is `BfabricAPIEngineType.SUDS`.
    """

    def __init__(
        self,
        config: BfabricClientConfig,
        auth: BfabricAuth | None,
        engine: BfabricAPIEngineType = BfabricAPIEngineType.SUDS,
    ) -> None:
        self.query_counter = 0
        self._config = config
        self._auth = auth
        self._engine_type = engine
        self._log_version_message()

    @cached_property
    def _engine(self) -> EngineSUDS | EngineZeep:
        if self._engine_type == BfabricAPIEngineType.SUDS:
            return EngineSUDS(base_url=self._config.base_url)
        elif self._engine_type == BfabricAPIEngineType.ZEEP:
            return EngineZeep(base_url=self._config.base_url)
        else:
            raise ValueError(f"Unexpected engine type: {self._engine_type}")

    @classmethod
    def from_config(
        cls,
        config_env: str | None = None,
        config_path: str | None = None,
        auth: BfabricAuth | Literal["config"] | None = "config",
        engine: BfabricAPIEngineType = BfabricAPIEngineType.SUDS,
    ) -> Bfabric:
        """Returns a new Bfabric instance, configured with the user configuration file.
        If the `config_env` is specified then it will be used, if it is not specified the default environment will be
        determined by checking the following in order (picking the first one that is found):
        - The `BFABRICPY_CONFIG_ENV` environment variable
        - The `default_config` field in the config file "GENERAL" section
        :param config_env: Configuration environment to use. If not given, it is deduced as described above.
        :param config_path: Path to the config file, in case it is different from default
        :param auth: Authentication to use. If "config" is given, the authentication will be read from the config file.
            If it is set to None, no authentication will be used.
        :param engine: Engine to use for the API.
        """
        config, auth_config = get_system_auth(config_env=config_env, config_path=config_path)
        auth_used: BfabricAuth | None = auth_config if auth == "config" else auth
        return cls(config, auth_used, engine=engine)

    @classmethod
    def from_token(
        cls,
        token: str,
        config_env: str | None = None,
        config_path: str | None = None,
        engine: BfabricAPIEngineType = BfabricAPIEngineType.SUDS,
    ) -> Bfabric:
        """Returns a new Bfabric instance, configured with the user configuration file and the provided token.
        Any authentication in the configuration file will be ignored, but it will be used to determine the correct
        B-Fabric instance.
        :param token: the token to use for authentication
        :param config_env: the config environment to use (if not specified, see `from_config`)
        :param config_path: the path to the config file (if not specified, see `from_config`)
        :param engine: the engine to use for the API.
        """
        config, _ = get_system_auth(config_env=config_env, config_path=config_path)
        token_data = get_token_data(client_config=config, token=token)
        auth = BfabricAuth(login=token_data.user, password=token_data.user_ws_password)
        return cls(config, auth, engine=engine)

    @property
    def config(self) -> BfabricClientConfig:
        """Returns the config object."""
        return self._config

    @property
    def auth(self) -> BfabricAuth:
        """Returns the auth object.
        :raises ValueError: If authentication is not available
        """
        if self._auth is None:
            raise ValueError("Authentication not available")
        return self._auth

    @contextmanager
    def with_auth(self, auth: BfabricAuth) -> Generator[None, None, None]:
        """Context manager that temporarily (within the scope of the context) sets the authentication for
        the Bfabric object to the provided value. This is useful when authenticating multiple users, to avoid accidental
        use of the wrong credentials.
        """
        old_auth = self._auth
        self._auth = auth
        try:
            yield
        finally:
            self._auth = old_auth

    def read(
        self,
        endpoint: str,
        obj: dict[str, Any],
        max_results: int | None = 100,
        offset: int = 0,
        check: bool = True,
        return_id_only: bool = False,
    ) -> ResultContainer:
        """Reads from the specified endpoint matching all specified attributes in `obj`.
        By setting `max_results` it is possible to change the number of results that are returned.
        :param endpoint: the endpoint to read from, e.g. "sample"
        :param obj: a dictionary containing the query, for every field multiple possible values can be provided, the
            final query requires the condition for each field to be met
        :param max_results: cap on the number of results to query. The code will keep reading pages until all pages
           are read or expected number of results has been reached. If None, load all available pages.
           NOTE: max_results will be rounded upwards to the nearest multiple of BFABRIC_QUERY_LIMIT, because results
           come in blocks, and there is little overhead to providing results over integer number of pages.
        :param offset: the number of elements to skip before starting to return results (useful for pagination, default
              is 0 which means no skipping)
        :param check: whether to raise an error if the response is not successful
        :param return_id_only: whether to return only the ids of the found objects
        :return: List of responses, packaged in the results container
        """
        # Get the first page.
        logger.debug(f"Reading from endpoint {repr(endpoint)} with query {repr(obj)}")
        results = self._engine.read(
            endpoint=endpoint,
            obj=obj,
            auth=self.auth,
            page=1,
            return_id_only=return_id_only,
        )
        n_available_pages = results.total_pages_api
        if not n_available_pages:
            if check:
                results.assert_success()
            return results.get_first_n_results(max_results)

        # Get results from other pages as well, if need be
        requested_pages, initial_offset = compute_requested_pages(
            n_page_total=n_available_pages,
            n_item_per_page=BFABRIC_QUERY_LIMIT,
            n_item_offset=offset,
            n_item_return_max=max_results,
        )
        logger.debug(f"Requested pages: {requested_pages}")

        # NOTE: Page numbering starts at 1
        response_items: list[dict[str, Any]] = []
        errors = results.errors
        page_offset = initial_offset
        for i_iter, i_page in enumerate(requested_pages):
            if not (i_iter == 0 and i_page == 1):
                logger.debug(f"Reading page {i_page} of {n_available_pages}")
                results = self._engine.read(
                    endpoint=endpoint,
                    obj=obj,
                    auth=self.auth,
                    page=i_page,
                    return_id_only=return_id_only,
                )
                errors += results.errors

            response_items += results[page_offset:]
            page_offset = 0

        result = ResultContainer(response_items, total_pages_api=n_available_pages, errors=errors)
        if check:
            result.assert_success()
        return result.get_first_n_results(max_results)

    def save(
        self,
        endpoint: str,
        obj: dict[str, Any],
        check: bool = True,
        method: str = "save",
    ) -> ResultContainer:
        """Saves the provided object to the specified endpoint.
        :param endpoint: the endpoint to save to, e.g. "sample"
        :param obj: the object to save
        :param check: whether to raise an error if the response is not successful
        :param method: the method to use for saving, generally "save", but in some cases e.g. "checkandinsert" is more
            appropriate to be used instead.
        :return a ResultContainer describing the saved object if successful
        """
        results = self._engine.save(endpoint=endpoint, obj=obj, auth=self.auth, method=method)
        if check:
            results.assert_success()
        return results

    def delete(self, endpoint: str, id: int | list[int], check: bool = True) -> ResultContainer:
        """Deletes the object with the specified ID from the specified endpoint.
        :param endpoint: the endpoint to delete from, e.g. "sample"
        :param id: the ID of the object to delete
        :param check: whether to raise an error if the response is not successful
        :return a ResultContainer describing the deleted object if successful
        """
        results = self._engine.delete(endpoint=endpoint, id=id, auth=self.auth)
        if check:
            results.assert_success()
        return results

    def exists(
        self,
        endpoint: str,
        key: str,
        value: int | str,
        query: dict[str, Any] | None = None,
        check: bool = True,
    ) -> bool:
        """Returns whether an object with the specified key-value pair exists in the specified endpoint.
        Further conditions can be specified in the query.
        :param endpoint: the endpoint to check, e.g. "sample"
        :param key: the key to check, e.g. "id"
        :param value: the value to check, e.g. 123
        :param query: additional query conditions (optional)
        :param check: whether to raise an error if the response is not successful
        """
        query = query or {}
        results = self.read(
            endpoint=endpoint,
            obj={**query, key: value},
            max_results=1,
            check=check,
            return_id_only=True,
        )
        return len(results) > 0

    def upload_resource(
        self, resource_name: str, content: bytes, workunit_id: int, check: bool = True
    ) -> ResultContainer:
        """Uploads a resource to B-Fabric, only intended for relatively small files that will be tracked by B-Fabric
        and not one of the dedicated experimental data stores.
        :param resource_name: the name of the resource to create (the same name can only exist once per workunit)
        :param content: the content of the resource as bytes
        :param workunit_id: the workunit ID to which the resource belongs
        :param check: whether to check for errors in the response
        """
        content_encoded = base64.b64encode(content).decode()
        return self.save(
            endpoint="resource",
            obj={
                "base64": content_encoded,
                "name": resource_name,
                "description": "base64 encoded file",
                "workunitid": workunit_id,
            },
            check=check,
        )

    def _get_version_message(self) -> tuple[str, str]:
        """Returns the version message as a string."""
        package_version = importlib.metadata.version("bfabric")
        year = datetime.now().year
        engine_name = self._engine.__class__.__name__
        base_url = self.config.base_url
        user_name = f"U={self._auth.login if self._auth else None}"
        python_version = f"PY={sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}"
        return (
            f"bfabricPy v{package_version} ({engine_name}, {base_url}, {user_name}, {python_version})",
            f"Copyright (C) 2014-{year} Functional Genomics Center Zurich",
        )

    def _log_version_message(self) -> None:
        """Logs the version message describing bfabricpy version, engine and base url."""
        console = Console(highlighter=HostnameHighlighter(), theme=DEFAULT_THEME)
        for line in self._get_version_message():
            with console.capture() as capture:
                console.print(line, style="bright_yellow", end="")
            logger.info(capture.get())

    def __repr__(self) -> str:
        return f"Bfabric(config={repr(self.config)}, auth={repr(self.auth)}, engine={self._engine})"

    __str__ = __repr__

    def __getstate__(self) -> dict[str, Any]:
        return {
            "config": self._config,
            "auth": self._auth,
            "engine_type": self._engine_type,
            "query_counter": self.query_counter,
        }

    def __setstate__(self, state: dict[str, Any]) -> None:
        self._config = state["config"]
        self._auth = state["auth"]
        self._engine_type = state["engine_type"]
        self.query_counter = state["query_counter"]


def get_system_auth(
    login: str | None = None,
    password: str | None = None,
    base_url: str | None = None,
    config_path: str | None = None,
    config_env: str | None = None,
    optional_auth: bool = True,
    verbose: bool = False,
) -> tuple[BfabricClientConfig, BfabricAuth | None]:
    """
    :param login:           Login string for overriding config file
    :param password:        Password for overriding config file
    :param base_url:        Base server url for overriding config file
    :param config_path:     Path to the config file, in case it is different from default
    :param config_env:      Which config environment to use. Can also specify via environment variable or use
       default in the config file (at your own risk)
    :param optional_auth:   Whether authentication is optional. If yes, missing authentication will be ignored,
       otherwise an exception will be raised
    :param verbose:         Verbosity (TODO: resolve potential redundancy with logger)
    """
    resolved_path = Path(config_path or "~/.bfabricpy.yml").expanduser()

    # Use the provided config data from arguments instead of the file
    if not resolved_path.is_file():
        if config_path:
            # NOTE: If user explicitly specifies a path to a wrong config file, this has to be an exception
            raise OSError(f"Explicitly specified config file does not exist: {resolved_path}")
        # TODO: Convert to log
        print(f"Warning: could not find the config file in the default location: {resolved_path}")
        config = BfabricClientConfig(base_url=base_url)
        auth = None if login is None or password is None else BfabricAuth(login=login, password=password)

    # Load config from file, override some of the fields with the provided ones
    else:
        config, auth = read_config(resolved_path, config_env=config_env)
        config = config.copy_with(base_url=base_url)
        if (login is not None) and (password is not None):
            auth = BfabricAuth(login=login, password=password)
        elif (login is None) and (password is None):
            pass
        else:
            raise OSError("Must provide both username and password, or neither.")

    if not config.base_url:
        raise ValueError("base_url missing")
    if not optional_auth and (not auth or not auth.login or not auth.password):
        raise ValueError("Authentication not initialized but required")

    if verbose:
        pprint(config)

    return config, auth



from __future__ import annotations

from loguru import logger

"""
Interface to the SLURM (Simple Linux Utility for Resources Management) resource manager and job scheduler

2020-09-28
Maria d'Errico
Christian Panse
"""

# Copyright (C) 2011, 2012 ETH Zurich and University of Zurich. All rights reserved.
# Copyright (C) 2014 Functional Genomics Center Zurich ETHZ|UZH. All rights reserved.
#
# Authors:
#   Christian Panse <cp@fgcz.ethz.ch>
#   Maria d'Errico <maria.derrico@fgcz.ethz.ch>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from pathlib import Path
import os
import subprocess


class SLURM:
    """Wrapper for SLURM, providing a Python interface to `sbatch`.

    The `slurm_root` variable will be passed as `SLURMROOT` to the environment, when submitting the script, and is an
    important parameter which needs to be set correctly for our scripts to function properly.
    """

    def __init__(self, slurm_root: str | Path = "/usr/") -> None:
        self._slurm_root = Path(slurm_root)
        self._sbatch_bin = self._slurm_root / "bin/sbatch"

    def sbatch(self, script: str | Path) -> tuple[str, str] | None:
        """Submits the script to SLURM using `sbatch`.
        If successful, returns a tuple with the stdout and stderr of the submission.
        """
        script = Path(script)
        if not script.is_file():
            logger.error(f"Script not found: {script}")
            return
        if not self._sbatch_bin.is_file():
            logger.error(f"sbatch binary not found: {self._sbatch_bin}")
            return

        env = os.environ | {"SLURMROOT": self._slurm_root}
        result = subprocess.run(
            [str(self._sbatch_bin), str(script)],
            env=env,
            check=True,
            shell=False,
            capture_output=True,
            encoding="utf-8",
        )
        # TODO the code initially had a TODO to write these two to a file, in general I think the logs of the squeue
        #      are currently not written to a file at all.
        return result.stdout, result.stderr
